""" Web interface routes and page generator. """

from flask import render_template, url_for, redirect, send_from_directory, abort, request, session
from flask_discord.utils import requires_authorization
from flask_discord.exceptions import Unauthorized
from time import time, sleep
from pytz import timezone as tz
from pytz.exceptions import UnknownTimeZoneError
from discord import Embed, Color
from discord.utils import get
from aya import logger, redis, nitori, leaderboards, loop, invite_url, system_administrators, operators, \
    default_prefix, config, leaderboard_status, mod_apply_status
from aya.frontend import app, discord, login, forms
from aya.utils import web_related, customize, experience
from aya.utils.leaderboard import leaderboard_builder


@login.user_loader
def load_user():
    return discord.fetch_user()


# Begin application

@app.route('/touhou')
def touhou():
    return render_template("pages/pre-auth/touhou.html")


@app.route('/leaderboard/<int:gid>/ranks')
def ranks(gid):
    if not len(str(gid)) == 18:
        abort(404)
        return
    if not redis.exists(f"exp.{gid}"):
        abort(404)
        return
    guild = nitori.get_guild(int(gid))
    rank_list = customize.get_ranks(gid)
    rank_info = []
    if rank_list is not None:
        for rank in sorted(rank_list, reverse=False):
            try:
                rank_info_entry = dict(level=int(rank), exp=experience.calculate_exp(int(rank)),
                                       name=get(guild.roles, id=int(rank_list[rank])).name)
                rank_info.append(rank_info_entry)
            except AttributeError:
                customize.remove_rank(int(gid), int(rank))
    return render_template('pages/leaderboard/ranks.html',
                           guild_name=guild.name,
                           guild_icon=guild.icon_url,
                           guild_id=guild.id,
                           rank_info=rank_info,
                           leaderboard_url=f'/leaderboard/{gid}',
                           ranks_url=f'/leaderboard/{gid}/ranks',
                           login=url_for('login'))


@app.route('/leaderboard/<int:gid>')
def leaderboard(gid):
    if not len(str(gid)) == 18:
        return abort(404)
    if not redis.exists(f"exp.{gid}"):
        return abort(404)
    if gid not in leaderboards:
        leaderboards.update({gid: {"users": [], "updated": 0}})
    users = leaderboards[gid]["users"]
    guild = nitori.get_guild(int(gid))
    if config.get("cache_request_based", False):
        if int(time()) - leaderboards[gid].get("updated", 0) >= int(config.get("leaderboard_refresh", "3600")):
            leaderboards[gid]["updated"] = int(time())
            experiences = experience.get_exps(gid)
            levels = sorted(experiences, key=experiences.get, reverse=True)
            loop.create_task(leaderboard_builder(levels, gid))
            while leaderboard_status.get(gid) is False:
                sleep(.1)
            return redirect(f"/leaderboard/{gid}")
    return render_template('pages/leaderboard/leaderboard.html',
                           users=users,
                           guild_name=guild.name,
                           guild_icon=guild.icon_url,
                           guild_id=guild.id,
                           leaderboard_url=f'/leaderboard/{gid}',
                           ranks_url=f'/leaderboard/{gid}/ranks',
                           login=url_for('login'))


@app.route('/apply/<int:gid>')
@requires_authorization
def apply(gid):
    if not len(str(gid)) == 18:
        return abort(404)
    user = discord.fetch_user()
    if mod_apply_status.get(gid, None) is not None:
        if int(time()) - mod_apply_status[gid].get(user.id, 0) < int(config.get("mod_apply_cooldown", "259200")):
            return abort(403)
    guild_list = discord.fetch_guilds()
    guild_thing = []
    for guild_obj in guild_list:
        if guild_obj.id == gid:
            guild_thing.append(guild_obj)
    if len(guild_thing) == 0:
        return abort(404)
    if customize.get_log_channel(gid) is None:
        return abort(404)
    guild = guild_thing[0]
    return render_template('pages/pre-auth/apply.html',
                           guild=guild,
                           submit_url=f"{request.path}/submit")


@app.route('/apply/<int:gid>/submit')
@requires_authorization
def apply_submit(gid):
    if not len(str(gid)) == 18:
        return abort(404)
    user = discord.fetch_user()
    if mod_apply_status.get(gid, None) is not None:
        if int(time()) - mod_apply_status[gid].get(user.id, 0) < int(config.get("mod_apply_cooldown", "259200")):
            return abort(403)
    guild_list = discord.fetch_guilds()
    guild_thing = []
    log_channel = customize.get_log_channel(gid)
    for guild_obj in guild_list:
        if guild_obj.id == gid:
            guild_thing.append(guild_obj)
    if len(guild_thing) == 0:
        return abort(404)
    if log_channel is None:
        return abort(404)
    voice_chat = request.args.get("voice-chat")
    age_range = request.args.get("age-range")
    past_experience = request.args.get("past-experience")
    timezone = request.args.get("timezone")
    form_whatever = request.args.get("form-whatever")
    if voice_chat == "NO_MIC":
        voice_chat = "Does not have microphone."
    elif voice_chat == "NO_VOICE":
        voice_chat = "Has microphone, does not want to use voice chat."
    elif voice_chat == "YES":
        voice_chat = "Able to voice chat."
    else:
        return abort(400)
    if past_experience == "NONE":
        past_experience = "No experience."
    elif past_experience == "MODERATE":
        past_experience = "Moderate experience."
    elif past_experience == "EXPERIENCED":
        past_experience = "Highly experienced."
    else:
        return abort(400)
    if age_range == "ADULT":
        age_range = "Above 18 years of age."
    elif age_range == "NOT_ADULT":
        age_range = "Below 18 years of age."
    else:
        return abort(400)
    if not len(form_whatever) <= 256 or not form_whatever.isascii():
        return abort(400)
    try:
        timezone = tz(timezone)
    except UnknownTimeZoneError:
        return abort(400)
    if mod_apply_status.get(gid, None) is None:
        mod_apply_status[gid] = {}
    mod_apply_status[gid][user.id] = int(time())
    content = Embed(title="Staff Application",
                    description="Newly received staff application.",
                    color=Color.blue())
    content.set_author(name=f"{user.name}#{user.discriminator}",
                       icon_url=user.avatar_url)
    content.add_field(name="Timezone",
                      value=str(timezone),
                      inline=False)
    content.add_field(name="Electronic Mail",
                      value=user.email,
                      inline=False)
    content.add_field(name="Voice Chat",
                      value=voice_chat,
                      inline=False)
    content.add_field(name="Moderation Experience",
                      value=past_experience,
                      inline=False)
    content.add_field(name="Age Range",
                      value=age_range,
                      inline=False)
    if form_whatever is not None and form_whatever != "":
        content.add_field(name="Additional Content",
                          value=form_whatever,
                          inline=False)
    content.set_footer(text=f"User: {user.id}")
    loop.create_task(log_channel.send(embed=content))
    return render_template("pages/pre-auth/apply_done.html")


@app.route('/admin/guild/<int:gid>', methods=['GET', 'POST'])
@requires_authorization
def guild_configurator(gid):
    if not len(str(gid)) == 18:
        abort(404)
        return
    if not any(guild.id == gid for guild in nitori.guilds):
        abort(404)
        return
    user = discord.fetch_user()
    guild_list = discord.fetch_guilds()
    guilds = web_related.get_guild_array(guild_list, user)
    guild = web_related.get_guild_object(gid, guilds, user)
    if not guild:
        return abort(403)
    current_message = dict(prefix=customize.get_prefix(gid) or default_prefix,
                           levelup=customize.get_message(gid, "levelup") or config.get("default_levelup_message",
                                                                                       "Congratulations {user} on "
                                                                                       "reaching level {level}."),
                           notification_channel=int(customize.get_message(gid, "notification_channel")) or "",
                           welcome_message=customize.get_message(gid, "welcome_message") or "",
                           welcome_image_url=customize.get_message(gid, "welcome_image_url") or "",
                           goodbye_message=customize.get_message(gid, "goodbye_message") or "",
                           goodbye_image_url=customize.get_message(gid, "goodbye_image_url") or "")
    current_leveling = dict(enabled=True)
    message_form = forms.MessageModifyForm(data=current_message, formdata=request.form)
    leveling_form = forms.LevelingSystemForm(data=current_leveling, formdata=request.form)
    if request.method == 'POST':
        if message_form.validate_on_submit():
            customize.set_prefix(gid, request.form.get('prefix', '', type=str))
            customize.set_message(gid, "levelup", request.form.get('levelup', '', type=str))
            customize.set_message(gid, "notification_channel", request.form.get('notification_channel', '', type=str))
            customize.set_message(gid, "welcome_message", request.form.get('welcome_message', '', type=str))
            customize.set_message(gid, "welcome_image_url", request.form.get('welcome_image_url', '', type=str))
            customize.set_message(gid, "goodbye_message", request.form.get('goodbye_message', '', type=str))
            customize.set_message(gid, "goodbye_image_url", request.form.get('goodbye_image_url', '', type=str))
        else:
            return redirect(url_for('admin'))
    if user.avatar_hash is not None:
        profile_picture = f"https://cdn.discordapp.com/avatars/{user.id}/{user.avatar_hash}.png"
    else:
        profile_picture = "https://cdn.discordapp.com/embed/avatars/1.png"
    return render_template('pages/admin/guild.html',
                           logout=url_for("logout"),
                           admin=url_for("admin"),
                           leaderboard=url_for("leaderboard", gid=gid),
                           username=user.name,
                           profile_picture=profile_picture,
                           invite=invite_url[0],
                           guild=guild,
                           guilds=guilds,
                           message_form=message_form,
                           leveling_form=leveling_form)


# End application


# Begin static files

@app.route('/favicon.ico')
def favicon():
    return send_from_directory('static', 'favicon.ico')


# End static files


@app.route("/logout/")
@requires_authorization
def logout():
    discord.revoke()
    return redirect(url_for("index"))


@app.route("/login/")
def login():
    if request.args.get("next") is not None:
        session["next"] = request.args.get("next")
    if discord.authorized:
        return redirect(url_for("admin") if request.args.get("next") is None else request.args.get("next"))
    return discord.create_session(scope=["identify", "email", "guilds"])


# Begin system structure


@app.route("/callback/")
def callback():
    try:
        discord.callback()
    except BaseException:
        return redirect(url_for("index"))
    return redirect(session.get("next", url_for("admin")))


@app.route("/admin/")
def admin():
    if not discord.authorized:
        return redirect(url_for("login"))
    try:
        user = discord.fetch_user()
    except Unauthorized:
        discord.revoke()
        return redirect(url_for("login"))
    if user.avatar_hash is not None:
        profile_picture = f"https://cdn.discordapp.com/avatars/{user.id}/{user.avatar_hash}.png"
    else:
        profile_picture = "https://cdn.discordapp.com/embed/avatars/1.png"
    guild_list = discord.fetch_guilds()
    guilds = web_related.get_guild_array(guild_list, user)
    return render_template('pages/admin/overview.html',
                           username=user.name,
                           email=user.email,
                           discriminator=user.discriminator,
                           userid=user.id,
                           profile_picture=profile_picture,
                           guilds=guilds,
                           invite=invite_url[0],
                           guilds_length=len(guild_list),
                           type="System Administrator" if user.id in system_administrators
                           else "Operator" if user.id in operators else "User",
                           mfa="Enabled" if user.mfa_enabled else "Disabled",
                           verified_mail="Yes" if user.verified else "No",
                           logout=url_for("logout"),
                           admin=url_for("admin"))


@app.route("/ja")
def index_ja():
    return render_template('pages/pre-auth/homepage.ja.html',
                           login=url_for("login"),
                           admin=url_for("admin"),
                           english=url_for("index"),
                           invite=invite_url[0])


@app.route("/")
def index():
    return render_template('pages/pre-auth/homepage.html',
                           login=url_for("login"),
                           admin=url_for("admin"),
                           japanese=url_for("index_ja"),
                           invite=invite_url[0])


@app.errorhandler(400)
def permission_denied(exception):
    logger.debug(exception)
    return render_template("pages/pre-auth/error.html",
                           error="Bad Request",
                           message="Your browser made a malformed request, please try again."), 400


@app.errorhandler(403)
def permission_denied(exception):
    logger.debug(exception)
    return render_template("pages/pre-auth/error.html",
                           error="Permission Denied",
                           message="You do not have access permission to this page."), 403


@app.errorhandler(404)
def no_such_file_or_directory(exception):
    logger.debug(exception)
    return render_template("pages/pre-auth/error.html",
                           error="Not Found",
                           message="The requested route does not exist on the Nitori system."), 404


@app.errorhandler(500)
def internal_server_error(exception):
    logger.debug(exception)
    return render_template("pages/pre-auth/error.html",
                           error="Internal Server Error",
                           message="The Nitori system has encountered an exception and is unable to process this "
                                   "request."), 500


@app.errorhandler(Unauthorized)
def discord_unauthorized(exception):
    logger.debug(exception)
    return redirect(url_for("login", next=request.path))

# End system structure
