""" Configurator for per-guild settings. """

from typing import Union
from emoji import UNICODE_EMOJI
from discord import Embed, Role, PartialEmoji, Color, TextChannel
from discord.utils import get
from discord.ext.commands import Cog, group, guild_only, command
from aya import nitori, redis_status, logger, default_prefix, config
from aya.utils import customize, permissions, formatter, localization


class Customization(Cog):
    """ conf_desc """

    @command(aliases=["lang", "locale"])
    @guild_only()
    @permissions.has_permissions(administrator=True)
    async def language(self, context, lang: str = None):
        """ conf_help_language """
        if not lang:
            avail_locale = ""
            for locale in localization.locales:
                avail_locale += f"{localization.locales[locale]['lang']} ({locale})\n"
            embed = Embed(title=localization.get_string("conf_locale_title", context),
                          description=localization.get_string("conf_locale_desc", context),
                          color=Color.blue())
            embed.add_field(name=localization.get_string("conf_locale_current", context),
                            value=localization.get_string("lang", context))
            embed.add_field(name=localization.get_string("conf_locale_avail", context),
                            value=avail_locale)
            return await context.send(embed=embed)
        elif lang in localization.locales:
            customize.set_language(context.guild.id, lang)
            await context.send(localization.get_string("conf_locale_set", context).format(
                locale=localization.get_string('lang', context)))
        else:
            await context.send(localization.get_string("invalid_argument", context))

    @group()
    @guild_only()
    @permissions.has_permissions(administrator=True)
    async def message(self, context):
        """ conf_help_message """
        gid = context.guild.id
        if context.invoked_subcommand is None:
            embed = Embed(title=localization.get_string("conf_title", context),
                          description=localization.get_string("conf_desc", context))
            guild_prefix = customize.get_prefix(gid)
            guild_levelup = customize.get_message(gid, "levelup")
            guild_welcome_message = customize.get_message(gid, "welcome_message")
            guild_notification_channel = customize.get_message(gid, "notification_channel")
            guild_welcome_image_url = customize.get_message(gid, "welcome_image_url")
            guild_goodbye_message = customize.get_message(gid, "goodbye_message")
            guild_goodbye_image_url = customize.get_message(gid, "goodbye_image_url")
            embed.add_field(name="prefix",
                            value=guild_prefix
                            if guild_prefix else
                            default_prefix,
                            inline=False)
            embed.add_field(name="levelup",
                            value=guild_levelup
                            if guild_levelup else
                            config.get("default_levelup_message",
                                       "Congratulations {user} on reaching level {level}."),
                            inline=False)
            embed.add_field(name="notification_channel",
                            value=guild_notification_channel
                            if guild_notification_channel else
                            localization.get_string("conf_disabled", context),
                            inline=False)
            embed.add_field(name="welcome_message",
                            value=guild_welcome_message
                            if guild_welcome_message else
                            f"{localization.get_string('conf_disabled', context)} "
                            "(Placeholders: {mention} for mention, {name} for display name)",
                            inline=False)
            embed.add_field(name="welcome_image_url",
                            value=guild_welcome_image_url
                            if guild_welcome_image_url else
                            localization.get_string("conf_disabled", context),
                            inline=False)
            embed.add_field(name="goodbye_message",
                            value=guild_goodbye_message
                            if guild_goodbye_message else
                            f"{localization.get_string('conf_disabled', context)} "
                            "(Placeholders: {mention} for mention, {name} for display name)",
                            inline=False)
            embed.add_field(name="goodbye_image_url",
                            value=guild_goodbye_image_url
                            if guild_goodbye_image_url else
                            localization.get_string("conf_disabled", context),
                            inline=False)
            await context.send(embed=embed)

    @message.command(name="levelup")
    async def levelup(self, context, *args):
        """ conf_message_help_levelup """
        if not args:
            customize.reset_message(context.guild.id, "levelup")
            return await context.send(localization.get_string("conf_lvlup_msg_reset", context))
        content = ' '.join(args)
        if len(content) >= 1024:
            return await context.send(localization.get_string("conf_msg_too_long", context))
        customize.set_message(context.guild.id, "levelup", content)
        await context.send(localization.get_string("conf_lvlup_msg_set", context).format(content=content))

    @message.command(name="prefix")
    async def prefix(self, context, content: str = None):
        """ conf_message_help_prefix """
        if not content:
            customize.reset_prefix(context.guild.id)
            return await context.send(localization.get_string("conf_prefix_reset", context))
        if len(content) > 16:
            return await context.send(localization.get_string("conf_msg_too_long", context))
        if content[-1].isalpha() or content[-1].isdigit():
            content += " "
        customize.set_prefix(context.guild.id, content)
        await context.send(localization.get_string("conf_prefix_set", context).format(content=content))

    @message.command(name="notification_channel")
    async def notification_channel(self, context, channel: TextChannel = None):
        """ conf_message_notification_channel """
        if channel is None:
            customize.reset_message(context.guild.id, "notification_channel")
            await context.send(localization.get_string("conf_notification_disable", context))
            return
        customize.set_message(context.guild.id, "notification_channel", str(channel.id))
        await context.send(localization.get_string("conf_notification_configure", context).format(channel=channel.name))

    @message.command(name="welcome_message")
    async def welcome_message(self, context, *args):
        """ conf_message_welcome_message """
        if not args:
            customize.reset_message(context.guild.id, "welcome_message")
            await context.send(localization.get_string("conf_welcome_disable", context))
            return
        content = ' '.join(args)
        if len(content) >= 1024:
            return await context.send(localization.get_string("conf_msg_too_long", context))
        customize.set_message(context.guild.id, "welcome_message", content)
        await context.send(localization.get_string("conf_welcome_configure", context).format(content=content))

    @message.command(name="welcome_image_url")
    async def welcome_image_url(self, context, *args):
        """ conf_message_welcome_message_url """
        if not args:
            customize.reset_message(context.guild.id, "welcome_image_url")
            await context.send(localization.get_string("conf_welcome_image_disable", context))
            return
        content = ' '.join(args)
        if len(content) >= 1024:
            await context.send(localization.get_string("conf_msg_too_long", context))
            return
        if formatter.url_validator.match(content) is None:
            return await context.send(localization.get_string("conf_not_http_url", context))
        customize.set_message(context.guild.id, "welcome_image_url", content)
        await context.send(localization.get_string("conf_welcome_image_configure", context).format(content=content))

    @message.command(name="goodbye_message")
    async def goodbye_message(self, context, *args):
        """ conf_message_goodbye_message """
        if not args:
            customize.reset_message(context.guild.id, "goodbye_message")
            await context.send(localization.get_string("conf_goodbye_disable", context))
            return
        content = ' '.join(args)
        if len(content) >= 1024:
            return await context.send(localization.get_string("conf_msg_too_long", context))
        customize.set_message(context.guild.id, "goodbye_message", content)
        await context.send(localization.get_string("conf_goodbye_configure", context).format(content=content))

    @message.command(name="goodbye_image_url")
    async def goodbye_image_url(self, context, *args):
        """ conf_message_goodbye_message_url """
        if not args:
            customize.reset_message(context.guild.id, "goodbye_image_url")
            await context.send(localization.get_string("conf_goodbye_image_disable", context))
            return
        content = ' '.join(args)
        if len(content) >= 1024:
            await context.send(localization.get_string("conf_msg_too_long", context))
            return
        if formatter.url_validator.match(content) is None:
            return await context.send(localization.get_string("conf_not_http_url", context))
        customize.set_message(context.guild.id, "goodbye_image_url", content)
        await context.send(localization.get_string("conf_goodbye_image_configure", context).format(content=content))

    @command()
    @guild_only()
    @permissions.has_permissions(administrator=True)
    async def setrank(self, context, level: int = None, role: Role = None):
        """ conf_help_setrank """
        gid = context.guild.id
        if level is None:
            ranks = customize.get_ranks(gid)
            try:
                prefix = customize.get_prefix(context.guild.id)
            except AttributeError:
                prefix = default_prefix
            if not prefix:
                prefix = default_prefix
            if not ranks:
                content = Embed(title=localization.get_string("conf_rank_title", context),
                                description=localization.get_string("conf_rank_desc_fresh", context).format(
                                    prefix=prefix))
                return await context.send(embed=content)
            content = Embed(title=localization.get_string("conf_rank_title", context),
                            description=localization.get_string("conf_rank_desc", context).format(prefix=prefix))
            for rank in sorted(ranks, reverse=False):
                try:
                    content.add_field(name=f"Level {str(rank)}",
                                      value=get(context.guild.roles,
                                                id=int(ranks[rank])).name,
                                      inline=False)
                except AttributeError:
                    customize.remove_rank(gid, int(rank))
            return await context.send(embed=content)
        if role is None:
            customize.remove_rank(gid, level)
            return await context.send(localization.get_string("conf_rank_remove", context).format(level=level))
        customize.set_rank(gid, level, role.id)
        await context.send(localization.get_string("conf_rank_add", context).format(role=role.name, level=level))

    @command(aliases=["expbl"])
    @guild_only()
    @permissions.has_permissions(administrator=True)
    async def expblacklist(self, context, action: str = None):
        """ conf_help_expbl """
        try:
            prefix = customize.get_prefix(context.guild.id)
        except AttributeError:
            prefix = default_prefix
        if not prefix:
            prefix = default_prefix
        gid = context.guild.id
        channel = context.channel.id
        if action == "toggle":
            if not customize.get_exp_bl(gid, channel):
                customize.set_exp_bl(gid, channel)
                return await context.send(localization.get_string("conf_expbl_on", context))
            else:
                customize.remove_exp_bl(gid, channel)
                return await context.send(localization.get_string("conf_expbl_off", context))
        else:
            content = Embed(title=localization.get_string("conf_expbl_title", context),
                            description=localization.get_string("conf_expbl_desc", context).format(prefix=prefix))
            blacklist = customize.get_exp_bls(gid)
            for channel_id in blacklist:
                try:
                    content.add_field(name=get(context.guild.channels,
                                               id=int(channel_id)).name,
                                      value=channel_id,
                                      inline=False)
                except AttributeError:
                    customize.remove_exp_bl(gid, channel_id)
            return await context.send(embed=content)

    @group()
    @guild_only()
    @permissions.has_permissions(administrator=True)
    async def roleassign(self, context):
        """ conf_help_roleassign """
        if context.invoked_subcommand is None:
            content = Embed(
                title="Role Assignment Manuals",
                description="Configure self-assigned roles.")
            content.add_field(name="create",
                              value="Create a page of assignable roles.",
                              inline=False)
            content.add_field(name="destroy",
                              value="Destroy a page of assignable roles.",
                              inline=False)
            content.add_field(name="list",
                              value="List all page of assignable roles.",
                              inline=False)
            content.add_field(name="name",
                              value="Rename a page of assignable roles.",
                              inline=False)
            content.add_field(name="generate",
                              value="Generate a page of assignable roles.",
                              inline=False)
            content.add_field(name="associate",
                              value="Associate an emote with a role.",
                              inline=False)
            content.add_field(name="disassociate",
                              value="Disassociate an emote with a role.",
                              inline=False)
            await context.send(embed=content)

    @roleassign.command(name="create")
    async def ra_create(self, context, *, name=None):
        if name is None:
            return await context.send("Please specify a name for the new page.")
        try:
            page = customize.create_assignable_role_page(context.guild.id, name)
        except ValueError:
            return await context.send("Page out of bounds, you can create at most 8 pages.")
        await context.send(f"Created `{name}` as page {page}.")

    @roleassign.command(name="destroy")
    async def ra_destroy(self, context, page: int = None):
        if page is None:
            return await context.send("Please specify a page to destroy.")
        try:
            customize.drop_assignable_role_page(context.guild.id, page)
        except ValueError:
            return await context.send("This page doesn't exist.")
        await context.send(f"Destroyed page {page}.")

    @roleassign.command(name="name")
    async def ra_name(self, context, page: int = None, *, name: str = None):
        if page is None:
            return await context.send("Please specify a page to name.")
        if page >= customize.get_assignable_role_pages(context.guild.id):
            return await context.send("This page doesn't exist.")
        if not name:
            return await context.send(f"The current name for page {page} is "
                                      f"`{customize.get_assignable_role_page_title(context.guild.id, page)}`.")
        customize.set_assignable_role_page_title(context.guild.id, page, name)
        await formatter.update_role_assign_message(context, context.guild.id, page)
        await context.send(f"Successfully renamed page {page}.")

    @roleassign.command(name="generate")
    async def ra_generate(self, context, page: int = None):
        if page is None:
            return await context.send("Please specify a page to generate.")
        if page >= customize.get_assignable_role_pages(context.guild.id):
            return await context.send("This page doesn't exist.")
        content, emotes = await formatter.role_assign_message(context, context.guild.id, page)
        message = await context.send(content)
        for emote in emotes:
            await message.add_reaction(emote)
        customize.set_assignable_role_message_info(context.guild.id, page, message.channel.id, message.id)

    @roleassign.command(name="associate")
    async def ra_associate(self, context, page: int = None, emote: Union[PartialEmoji, str] = None, *,
                           role: Role = None):
        if page is None:
            return await context.send("Please specify a page to perform this action on.")
        if emote is None:
            return await context.send("Please specify the emote to associate with.")
        if isinstance(emote, PartialEmoji):
            emote_id = emote.id
            if nitori.get_emoji(emote_id) is None:
                return await context.send(localization.get_string("conf_assign_no_access", context))
            if role is None:
                current_role_id = customize.get_assignable_role(context.guild.id, page, emote_id)
                if current_role_id is not None:
                    role_name = get(context.guild.roles,
                                    id=int(customize.get_assignable_roles(context.guild.id, page)[emote_id])).name
                    return await context.send(f"This emote is currently associated with role `{role_name}`.")
                else:
                    return await context.send("This emote is not associated with any role.")
            customize.set_assignable_role(context.guild.id, page, role.id, emote_id)
            if len(customize.get_assignable_roles(context.guild.id, page)) > 64:
                customize.remove_assignable_role(context.guild.id, page, emote_id)
                return await context.send("You have reached the maximum amount of assignable roles.")
            await formatter.update_role_assign_message(context, context.guild.id, page)
            return await context.send(
                localization.get_string("conf_assign_emote_add", context).format(role=role.name, emote=emote))
        elif isinstance(emote, str) and len(emote) == 1:
            if emote[0] in UNICODE_EMOJI:
                if role is None:
                    current_role_id = customize.get_assignable_role(context.guild.id, page, emote)
                    if current_role_id is not None:
                        role_name = get(context.guild.roles,
                                        id=int(customize.get_assignable_roles(context.guild.id, page)[emote])).name
                        return await context.send(f"This emote is currently associated with role `{role_name}`.")
                    else:
                        return await context.send("This emote is not associated with any role.")
                customize.set_assignable_role(context.guild.id, page, role.id, emote)
                if len(customize.get_assignable_roles(context.guild.id, page)) > 64:
                    customize.remove_assignable_role(context.guild.id, page, emote)
                    return await context.send("You have reached the maximum amount of assignable roles.")
                await formatter.update_role_assign_message(context, context.guild.id, page)
                return await context.send(
                    localization.get_string("conf_assign_emote_add", context).format(role=role.name, emote=emote))
            else:
                return await context.send(localization.get_string("conf_assign_emote_invalid", context))

    @roleassign.command(name="disassociate")
    async def ra_disassociate(self, context, page: int = None, emote: Union[PartialEmoji, str] = None):
        if page is None:
            return await context.send("Please specify a page to perform this action on.")
        if emote is None:
            return await context.send("Please specify the emote to associate with.")
        if isinstance(emote, PartialEmoji):
            emote_id = emote.id
            if nitori.get_emoji(emote_id) is None:
                return await context.send(localization.get_string("conf_assign_no_access", context))
            customize.remove_assignable_role(context.guild.id, page, emote_id)
            await formatter.update_role_assign_message(context, context.guild.id, page)
            return await context.send(
                localization.get_string("conf_assign_emote_clear", context).format(emote=emote))
        elif isinstance(emote, str) and len(emote) == 1:
            if emote[0] in UNICODE_EMOJI:
                customize.remove_assignable_role(context.guild.id, page, emote)
                await formatter.update_role_assign_message(context, context.guild.id, page)
                return await context.send(
                    localization.get_string("conf_assign_emote_clear", context).format(emote=emote))
            else:
                return await context.send(localization.get_string("conf_assign_emote_invalid", context))


def setup(client):
    if redis_status():
        client.add_cog(Customization())
    else:
        logger.warning("The per-guild configuration system is disabled due to a database error.")
