""" Event and error handler class. """

from os import getpid
from random import choice
from signal import SIGKILL
from psutil import Process
from traceback import format_tb
from discord import Game, Status, Forbidden
from discord.utils import get, oauth_url
from discord.ext.commands import Cog, command
from discord.ext.commands.errors import MissingRequiredArgument, BadArgument, \
    CommandInvokeError, CheckFailure, CommandOnCooldown, CommandNotFound, BotMissingPermissions, MissingPermissions
from aya import nitori, config, debug, logger, presence, redis_status, invite_url, loop, \
    leaderboard_status, default_prefix
from aya.utils import commit_config, permissions, customize, experience, leaderboard, cleanup, formatter, \
    localization, voice


async def send_command_help(context):
    await context.send(localization.get_string("invalid_argument", context))
    # TODO: Implement help sender!!!!


class Handler(Cog, command_attrs=dict(hidden=True)):
    """ handler_desc """

    @Cog.listener()
    async def on_command_error(self, context, exception):
        if isinstance(exception, MissingRequiredArgument) or isinstance(exception, BadArgument):
            await send_command_help(context)
        elif isinstance(exception, CommandInvokeError):
            exception = exception.original
            _traceback = format_tb(exception.__traceback__)
            _traceback = ''.join(_traceback)
            message = f"```plaintext\n" \
                      f"A fatal error occurred in the FreeNitori runtime.\n" \
                      f"Invoked by: \"{context.message.content}\"\n" \
                      f"Exception: {type(exception).__name__}: {exception}\n```" \
                      f"Traceback:\n```python\n{_traceback}\n" \
                      f"```"
            await context.send(localization.get_string("command_exception", context))
            if debug:
                await context.send(f"Error report:\n{message}")
        elif isinstance(exception, CheckFailure):
            await context.send(str(exception))
        elif isinstance(exception, CommandOnCooldown):
            await context.send(localization.get_string("cooldown_in_effect", context).format(
                retry_after=f"{exception.retry_after:.2f}"))
        elif isinstance(exception, CommandNotFound):
            seeking_message_list = ["Kappa",
                                    "What do you want?",
                                    "I like cucumbers.",
                                    "<:FeelsKappa:713635502741520384>"]
            guild_prefix = None
            if context.guild is not None:
                guild_prefix = customize.get_prefix(context.guild.id)
            if context.message.content.startswith("<@!") or context.message.content.startswith("<@") and \
                    context.message.mentions[0].id == nitori.user.id:
                await context.send(choice(seeking_message_list))
            else:
                await context.send(
                    f"This command does not exist! Issue `"
                    f"{guild_prefix if guild_prefix else default_prefix}man"
                    f"` for a list of command manuals.")
        elif isinstance(exception, MissingPermissions) or isinstance(exception, BotMissingPermissions) \
                or isinstance(exception, Forbidden):
            try:
                await context.send(localization.get_string("missing_permission", context))
            except BaseException:
                pass

    @Cog.listener()
    async def on_command(self, context):
        try:
            logger.info(f"{context.author}@\"{context.guild.name}\" > {context.message.clean_content}")
        except AttributeError:
            logger.info(f"{context.author}@Private > {context.message.clean_content}")

    @Cog.listener()
    async def on_message(self, context):
        if context.guild is None:
            return
        if not str(context.type) == "MessageType.default":
            return
        if not customize.get_slur_filter_status(context.guild.id):
            return
        for regex in formatter.slur_regex:
            for word_thingy in context.content.split():
                if regex.match(word_thingy):
                    try:
                        await context.delete()
                    except Forbidden:
                        return
                    return await context.channel.send(
                        f"{context.author.mention} Chat violation! Do not use racial slurs!")

    @Cog.listener()
    async def on_member_join(self, context):
        uid = context.id
        gid = context.guild.id
        if permissions.roles and config.get("rank", True) and redis_status() and not context.bot:
            exp = experience.get_exp(uid, gid)
            if not exp:
                experience.set_exp(uid, gid, 0)
                exp = 0
            level = experience.calculate_level(exp)
            ranks = customize.get_ranks(gid)
            for rank in ranks:
                try:
                    ranked_role = get(context.guild.roles, id=ranks[rank])
                except AttributeError:
                    customize.remove_rank(gid, int(rank))
                    continue
                try:
                    await context.remove_roles(ranked_role)
                except Forbidden:
                    pass
                if level >= int(rank):
                    try:
                        await context.add_roles(ranked_role)
                    except Forbidden:
                        pass
        if not context.bot:
            await formatter.greeter_message(context,
                                            channel_id=customize.get_message(gid, "notification_channel"),
                                            message=customize.get_message(gid, "welcome_message"),
                                            gid=gid,
                                            uid=uid,
                                            display_name=context.name,
                                            image_url=customize.get_message(gid, "welcome_image_url"))

    @Cog.listener()
    async def on_member_remove(self, context):
        if not context.bot:
            await formatter.greeter_message(context,
                                            channel_id=customize.get_message(context.guild.id, "notification_channel"),
                                            message=customize.get_message(context.guild.id, "goodbye_message"),
                                            gid=context.guild.id,
                                            uid=context.id,
                                            display_name=context.name,
                                            image_url=customize.get_message(context.guild.id, "goodbye_image_url"))

    @Cog.listener()
    async def on_voice_state_update(self, context, pre, post):
        gid = context.guild.id
        if context.bot:
            return
        if pre.channel is not None and post.channel is None:
            if gid in voice.session_pool:
                session = voice.get_session(gid)
                if pre.channel.id == session.client.channel.id and len(pre.channel.members) == 1:
                    await session.cleanup()
                    voice.session_pool.pop(gid)

    @Cog.listener()
    async def on_raw_reaction_add(self, context):
        guild = nitori.get_guild(context.guild_id)
        if context.member.bot:
            return
        channel_ids, message_ids = customize.get_assignable_role_messages(context.guild_id)
        if channel_ids is None or message_ids is None:
            return
        if context.channel_id not in channel_ids or context.message_id not in message_ids:
            return
        if not context.event_type == "REACTION_ADD":
            return
        assignable_roles = customize.get_assignable_roles(str(context.guild_id), message_ids[context.message_id])
        if context.emoji.is_unicode_emoji():
            emoji_string = str(context.emoji)
        elif context.emoji.is_custom_emoji():
            emoji_string = context.emoji.id
        else:
            return
        if emoji_string in assignable_roles:
            try:
                role = get(guild.roles, id=int(assignable_roles[emoji_string]))
            except AttributeError:
                return
            try:
                await context.member.add_roles(role)
            except Forbidden:
                channel = nitori.get_channel(context.channel_id)
                try:
                    await channel.send(localization.get_string("missing_permission", context))
                except Forbidden:
                    return

    @Cog.listener()
    async def on_raw_reaction_remove(self, context):
        guild = nitori.get_guild(context.guild_id)
        member = guild.get_member(context.user_id)
        channel_ids, message_ids = customize.get_assignable_role_messages(context.guild_id)
        if channel_ids is None or message_ids is None:
            return
        if member.bot:
            return
        if context.channel_id not in channel_ids or context.message_id not in message_ids:
            return
        if not context.event_type == "REACTION_REMOVE":
            return
        assignable_roles = customize.get_assignable_roles(str(context.guild_id), message_ids[context.message_id])
        if context.emoji.is_unicode_emoji():
            emoji_string = str(context.emoji)
        elif context.emoji.is_custom_emoji():
            emoji_string = context.emoji.id
        else:
            return
        if emoji_string in assignable_roles:
            try:
                role = get(guild.roles, id=int(assignable_roles[emoji_string]))
            except AttributeError:
                return
            try:
                await member.remove_roles(role)
            except Forbidden:
                channel = nitori.get_channel(context.channel_id)
                try:
                    await channel.send(localization.get_string("missing_permission", context))
                except Forbidden:
                    return

    @Cog.listener()
    async def on_guild_remove(self, context):
        customize.reset_guild(context.id)

    @Cog.listener()
    async def on_guild_channel_create(self, context):
        role = customize.get_mute_role(context.guild.id)
        if role is not None:
            try:
                await context.set_permissions(role, send_messages=False,
                                              read_message_history=False,
                                              read_messages=False)
            except Forbidden:
                pass

    @Cog.listener()
    async def on_ready(self):
        invite_url.append(f"{oauth_url(nitori.user.id)}&permissions=2146958847")
        if presence is None:
            logger.fatal("Please define the presence field in the configuration.")
            await nitori.close()
            Process(getpid()).send_signal(SIGKILL)
            return
        logger.info(f"Nitori has successfully connected to discord.")
        logger.info(f"User: {nitori.user} | Guilds: {len(nitori.guilds)}")
        await nitori.change_presence(activity=Game(type=0, name=presence.get("playing", None)),
                                     status=Status.online if not presence.get("idle", False)
                                     else Status.idle)
        if presence.get("dnd", False):
            await nitori.change_presence(activity=Game(type=0, name=presence.get("playing", None)),
                                         status=Status.dnd)
        if debug:
            await nitori.change_presence(activity=Game(type=0, name=presence.get("playing", None)),
                                         status=Status.idle)
        leaderboard_status.update({"startup": True})
        if not config.get("cache_request_based", False):
            leaderboard.periodical_reload()
        else:
            loop.create_task(leaderboard.rebuild_leaderboards())

    @command()
    @permissions.system_administrators_only()
    async def debug(self, context):
        """ handler_help_debug """
        if debug is True:
            commit_config.set_key("debug", False)
            await context.send(localization.get_string("handler_production_mode", context))
            await cleanup.clean_shutdown()
        else:
            commit_config.set_key("debug", True)
            await context.send(localization.get_string("handler_debug_mode", context))
            await cleanup.clean_shutdown()


def setup(client):
    client.add_cog(Handler())
