""" Moderation related utilities. """

from discord import Object as DiscordObject, Member, Embed, Forbidden
from discord.ext.commands import Cog, command, guild_only, BadArgument, cooldown, BucketType
from aya import config
from aya.utils import permissions, MemberID, localization, customize
from aya.utils.formatter import date_formatter, reason_formatter, action_formatter


class Moderation(Cog):
    """ mod_desc """

    @command()
    @guild_only()
    async def whois(self, context, *, user: Member = None):
        """ mod_help_whois """
        if user is None:
            user = context.author
        embed = Embed(color=user.top_role.color.value)
        embed.set_thumbnail(url=user.avatar_url)
        embed.add_field(name=localization.get_string("mod_username", context),
                        value=user.name,
                        inline=True)
        embed.add_field(name=localization.get_string("mod_nickname", context),
                        value=user.nick if hasattr(user, "nick") else localization.get_string("mod_not_configured",
                                                                                              context),
                        inline=True)
        embed.add_field(
            name=localization.get_string("mod_roles", context),
            value=', '.join([f"<@&{x.id}>" for x in user.roles if x is not context.guild.default_role]) if len(
                user.roles) > 1 else localization.get_string("mod_not_configured", context),
            inline=False
        )
        embed.add_field(name=localization.get_string("mod_register_date", context),
                        value=date_formatter(user.created_at),
                        inline=True)
        embed.add_field(name=localization.get_string("mod_join_date", context),
                        value=date_formatter(user.joined_at),
                        inline=True)
        await context.send(embed=embed)

    @command()
    async def pfp(self, context, *, user: Member = None):
        """ mod_help_pfp """
        if user is None:
            user = context.author
        embed = Embed(color=user.color,
                      description=localization.get_string("mod_profile_picture", context).format(user=user.name))
        embed.set_image(url=user.avatar_url_as(size=4096))
        await context.send(embed=embed)

    @pfp.error
    async def pfp_handler(self, context, exception):
        if isinstance(exception, BadArgument):
            await context.send(localization.get_string("mod_user_missing", context))

    @command()
    @guild_only()
    @permissions.has_permissions(kick_members=True)
    async def kick(self, context, member: Member, *, reason: str = None):
        """ mod_help_kick """
        try:
            await member.kick(reason=reason_formatter(context.author, reason))
            await context.send(action_formatter("kick"))
        except Forbidden:
            await context.send(localization.get_string("missing_permission", context))

    @command(aliases=["nick"])
    @guild_only()
    @permissions.has_permissions(manage_nicknames=True)
    async def nickname(self, context, member: Member, *, name: str = None):
        """ mod_help_nickname """
        try:
            await member.edit(nick=name,
                              reason=reason_formatter(context.author,
                                                      localization.get_string("mod_default_audit_message", context)))
            message = localization.get_string("mod_nickname_set", context).format(username=member.name, name=name)
            if name is None:
                message = localization.get_string("mod_nickname_reset", context).format(username=member.name)
            await context.send(message)
        except Forbidden:
            await context.send(localization.get_string("missing_permission", context))

    @command()
    @guild_only()
    @permissions.has_permissions(ban_members=True)
    async def ban(self, context, member: MemberID, *, reason: str = None):
        """ mod_help_ban """
        try:
            await context.guild.ban(DiscordObject(id=member),
                                    reason=reason_formatter(context.author, reason))
            await context.send(action_formatter("ban"))
        except Forbidden:
            await context.send(localization.get_string("missing_permission", context))

    @command(aliases=["pardon"])
    @guild_only()
    @permissions.has_permissions(ban_members=True)
    async def unban(self, context, member: MemberID, *, reason: str = None):
        """ mod_help_unban """
        try:
            await context.guild.unban(DiscordObject(id=member),
                                      reason=reason_formatter(context.author, reason))
            await context.send(action_formatter("pardon"))
        except Forbidden:
            await context.send(localization.get_string("missing_permission", context))

    @command()
    @guild_only()
    @permissions.has_permissions(administrator=True)
    async def slurfilter(self, context):
        """ mod_help_slurfilter """
        if customize.toggle_slur_filter_status(context.guild.id):
            return await context.send("Enabled racial slur filters.")
        else:
            return await context.send("Disabled racial slur filters.")

    @command(aliases=["prison", "containment"])
    @guild_only()
    @permissions.has_permissions(administrator=True)
    async def gulag(self, context):
        """ mod_help_gulag """
        if customize.get_gulag_channel(context.guild.id) == context.channel:
            customize.remove_gulag_channel(context.guild.id)
            role = customize.get_mute_role(context.guild.id)
            if role is not None:
                try:
                    await context.channel.set_permissions(role, send_messages=False,
                                                          read_message_history=False,
                                                          read_messages=False)
                except Forbidden:
                    pass
            await context.send("Gulag channel has been unset.")
        else:
            customize.set_gulag_channel(context.guild.id, context.channel)
            role = customize.get_mute_role(context.guild.id)
            if role is None:
                try:
                    role = await context.guild.create_role(name="Muted", reason="First time mute setup.")
                    customize.set_mute_role(context.guild.id, role)
                    for channel in context.guild.channels:
                        if channel == context.channel:
                            await channel.set_permissions(context.guild.default_role, send_messages=False,
                                                          read_message_history=False,
                                                          read_messages=False)
                            await channel.set_permissions(role, send_messages=True,
                                                          read_message_history=True,
                                                          read_messages=True)
                            continue
                        await channel.set_permissions(role, send_messages=False,
                                                      read_message_history=False,
                                                      read_messages=False)
                except Forbidden:
                    return await context.send(localization.get_string("missing_permission", context))
            else:
                try:
                    await context.channel.set_permissions(context.guild.default_role, send_messages=False,
                                                          read_message_history=False,
                                                          read_messages=False)
                    await context.channel.set_permissions(role, send_messages=True,
                                                          read_message_history=True,
                                                          read_messages=True)
                except Forbidden:
                    return await context.send(localization.get_string("missing_permission", context))
            await context.send("Gulag channel has been set to current channel.")

    @command()
    @guild_only()
    @permissions.has_permissions(ban_members=True)
    async def mute(self, context, member: Member, *, reason: str = None):
        """ mod_help_mute """
        role = customize.get_mute_role(context.guild.id)
        gulag = customize.get_gulag_channel(context.guild.id)
        if role is None:
            try:
                role = await context.guild.create_role(name="Muted", reason="First time mute setup.")
                customize.set_mute_role(context.guild.id, role)
                for channel in context.guild.channels:
                    if channel == gulag:
                        await channel.set_permissions(context.guild.default_role, send_messages=False,
                                                      read_message_history=False,
                                                      read_messages=False)
                        await channel.set_permissions(role, send_messages=True,
                                                      read_message_history=True,
                                                      read_messages=True)
                        continue
                    await channel.set_permissions(role, send_messages=False,
                                                  read_message_history=False,
                                                  read_messages=False)
            except Forbidden:
                return await context.send(localization.get_string("missing_permission", context))
        try:
            await member.add_roles(role,
                                   reason=reason_formatter(context.author, reason))
            await context.send(action_formatter("global mute"))
        except Forbidden:
            await context.send(localization.get_string("missing_permission", context))

    @command()
    @guild_only()
    @permissions.has_permissions(ban_members=True)
    async def unmute(self, context, member: Member, *, reason: str = None):
        """ mod_help_unmute """
        role = customize.get_mute_role(context.guild.id)
        if role is None:
            return await context.send(action_formatter("global unmute"))
        try:
            await member.remove_roles(role,
                                      reason=reason_formatter(context.author, reason))
            await context.send(action_formatter("global unmute"))
        except Forbidden:
            await context.send(localization.get_string("missing_permission", context))

    @command()
    @guild_only()
    async def apply(self, context):
        """ mod_help_apply """
        if customize.get_log_channel(context.guild.id) is None:
            return await context.send("Staff application is disabled.")
        content = Embed(title=localization.get_string("mod_staff_title", context),
                        description=localization.get_string("mod_staff_body", context).format(
                            link=f"[{localization.get_string('mod_staff_link', context)}]("
                                 f"{config.get('url', 'http://localhost:7777/')}apply/{context.guild.id})"))
        await context.send(embed=content)

    @command()
    @guild_only()
    @cooldown(rate=1, per=1800, type=BucketType.user)
    async def report(self, context, member: Member, *, reason: str = None):
        """ mod_help_report """
        log_channel = customize.get_log_channel(context.guild.id)
        if log_channel is None:
            return await context.send("Reports are currently disabled.")
        await context.message.delete()
        message_url = (await context.send(
            f"Submitted report against `{member.display_name}`, "
            f"please wait for the staff to get back to you.")).jump_url
        content = Embed(title="Report", color=context.author.color)
        content.set_author(name=str(context.author), icon_url=context.author.avatar_url_as(format='png', size=128))
        content.add_field(name="Chat Message",
                          value=message_url,
                          inline=False)
        content.add_field(name="Reporter",
                          value=context.author.mention,
                          inline=False)
        content.add_field(name="Offender",
                          value=member.mention,
                          inline=False)
        if reason is not None:
            content.add_field(name="Reason",
                              value=reason,
                              inline=False)
        content.set_footer(text=f"Channel: {context.channel.id} (#{context.channel.name}) User: {context.author.id}")
        await log_channel.send(embed=content)


def setup(client):
    client.add_cog(Moderation())
