""" Voice channel music playback. """

from discord import PCMVolumeTransformer, Permissions
from discord.ext.commands import Cog, command, guild_only
from aya import nitori, config, logger, system_administrators, operators
from aya.utils import voice, youtube, formatter, permissions, localization


class Music(Cog):
    """ music_desc """

    @command()
    @guild_only()
    async def join(self, context):
        """ music_help_join """
        try:
            voice_channel = context.author.voice.channel
        except AttributeError:
            await context.send(localization.get_string("music_not_in_channel", context))
            return False

        session = voice.get_session(context.guild.id, True)
        if session.client is None:
            session.client = await voice_channel.connect()
            await context.send(localization.get_string("music_connected", context).format(voice_channel=voice_channel))
            return True
        else:
            if not context.author.guild_permissions.move_members:
                return await context.send(localization.get_string("permission_denied", context))
            await session.client.move_to(voice_channel)
            await context.send(localization.get_string("music_relocated", context).format(voice_channel=voice_channel))
            return True

    @command(aliases=["disconnect"])
    @guild_only()
    @permissions.has_permissions(move_members=True)
    async def leave(self, context):
        """ music_help_leave """
        gid = context.guild.id
        if gid not in voice.session_pool:
            return await context.send(localization.get_string("music_not_connected", context))
        session = voice.get_session(gid)
        await session.cleanup()
        await context.send(
            localization.get_string("music_disconnected", context).format(voice_channel=session.client.channel.name))
        voice.session_pool.pop(gid)

    @command(aliases=["p"])
    @guild_only()
    async def play(self, context, *, url: str):
        """ music_help_play """
        session = voice.get_session(context.guild.id, True)
        if session.processing:
            return await context.send(localization.get_string("music_busy", context))
        if session.client is None:
            if not await context.invoke(self.join):
                return
        if "playlist?list=" in url or "&list=" in url:
            if not config.get("playlist", False):
                return await context.send(localization.get_string("music_playlist_disabled", context))
            await voice.playlist(context, url)
            return
        if session.client.is_playing():
            session.processing = True
            information = youtube.extract_info(url, playlist=False)
            if 'entries' in information:
                information = information['entries'][0]
                url = information['webpage_url']
            await session.songs.put(url)
            await context.send(localization.get_string("music_added", context).format(title=information['title']))
            session.queue_message = formatter.queue_formatter(session)
            session.processing = False
            return
        session.processing = True
        processing_notification = await context.send(localization.get_string("music_processing", context))
        if context.voice_client is not None:
            information = youtube.extract_info(url, playlist=False)
            if information is None:
                session.processing = False
                return await processing_notification.edit(
                    content=localization.get_string("music_not_found", context))
            if "entries" in information:
                playlist = True
                information = information["entries"][0]
                url = information['webpage_url']
            else:
                session.processing = False
                return await processing_notification.edit(content=localization.get_string("music_not_found", context))
            duration = information['duration']
            if duration > 3600 and not config.get("hour_video", False):
                session.processing = False
                return await processing_notification.edit(
                    content=localization.get_string("music_hour_video_disabled", context))
            if config.get("aria", False):
                session.audio = await voice.YoutubeSource.from_url(url, loop=nitori.loop, aria=True)
            else:
                session.audio = await voice.YoutubeSource.from_url(url, loop=nitori.loop, aria=False)
            audio_source = PCMVolumeTransformer(session.audio)
            session.audio.player = audio_source
            await processing_notification.delete()
            await context.send(embed=formatter.music_formatter(information, playlist))
            session.queue_message = formatter.queue_formatter(session)
            session.processing = False
            context.voice_client.play(audio_source, after=lambda expand: voice.next_song(context))
        else:
            await processing_notification.delete()
            await context.send(localization.get_string("music_not_in_channel", context))

    @command(aliases=["s", "next"])
    @guild_only()
    async def skip(self, context):
        """ music_help_skip """
        session = voice.get_session(context.guild.id, False)
        if session is None:
            return await context.send(localization.get_string("music_not_connected", context))
        target = context.message.author
        if not session.client.is_playing():
            return await context.send(localization.get_string("music_inactive", context))
        if context.channel.permissions_for(context.author).manage_messages or \
                context.author.id in system_administrators + operators:
            await context.send(
                localization.get_string("music_vote_override", context).format(name=context.message.author.name))
            session.skip_votes.clear()
            if session.client.is_playing():
                session.client.stop()
        elif target.id not in session.skip_votes:
            session.skip_votes.append(target.id)
            vote_amount = len(session.skip_votes)
            if vote_amount >= 3:
                await context.send(localization.get_string("music_vote_pass", context))
                session.skip_votes.clear()
                if session.client.is_playing():
                    session.client.stop()
            else:
                await context.send(
                    localization.get_string("music_vote_registered", context).format(vote_amount=vote_amount))
        else:
            await context.send(localization.get_string("music_already_voted", context))

    @command(aliases=["q", "np"])
    @guild_only()
    async def queue(self, context):
        """ music_help_queue """
        session = voice.get_session(context.guild.id, False)
        if session is None:
            return await context.send(localization.get_string("music_not_connected", context))
        await context.send(embed=session.queue_message)


def setup(client):
    if config.get("music", True):
        client.add_cog(Music())
    else:
        logger.warning("The music system is disabled explicitly in configuration.")
