""" Utilities to operate the Nitori system. """

from os import getpid
from io import BytesIO
from asyncio import sleep
from discord import File, Color, Embed, Member
from discord.ext.commands import Cog, command, guild_only, group
from psutil import Process, cpu_percent, virtual_memory, getloadavg
from aya import nitori, config, redis_status, invite_url, system_administrators, operators
from aya.utils import execute, permissions, experience, cleanup, formatter, localization


class System(Cog):
    """ sys_desc """

    @command()
    @permissions.system_administrators_only()
    async def reboot(self, context):
        """ sys_help_reboot """
        await context.send(localization.get_string("invoke_restart", context))
        await cleanup.clean_shutdown()

    @command()
    @permissions.system_administrators_only()
    async def reload(self, context):
        """ sys_help_reload """
        message = await context.send(localization.get_string("sched_reload", context))
        nitori.loop.create_task(cleanup.reload_internals(message))

    @group()
    @permissions.operator_only()
    async def plugin(self, context):
        """ sys_help_plugin """
        if context.invoked_subcommand is None:
            content = Embed(title=localization.get_string("pm_title", context),
                            description=localization.get_string("pm_desc", context))
            content.add_field(name="load",
                              value=localization.get_string("pm_load_desc", context),
                              inline=False)
            content.add_field(name="unload",
                              value=localization.get_string("pm_unload_desc", context),
                              inline=False)
            content.add_field(name="reload",
                              value=localization.get_string("pm_reload_desc", context),
                              inline=False)
            await context.send(embed=content)

    @plugin.command(name="load")
    @permissions.operator_only()
    async def plugin_load(self, context, name: str):
        """ sys_help_plugin_load """
        try:
            nitori.load_extension(f"plugins.{name}")
        except Exception:
            return await context.send(localization.get_string("pm_load_error", context).format(plugin=name))
        await context.send(localization.get_string("pm_load_success", context).format(plugin=name))

    @plugin.command(name="unload")
    @permissions.operator_only()
    async def plugin_unload(self, context, name: str):
        """ sys_help_plugin_unload """
        try:
            nitori.unload_extension(f"plugins.{name}")
        except Exception:
            return await context.send(localization.get_string("pm_unload_error", context).format(plugin=name))
        await context.send(localization.get_string("pm_unload_success", context).format(plugin=name))

    @plugin.command(name="reload")
    @permissions.operator_only()
    async def plugin_reload(self, context, name: str):
        """ sys_help_plugin_reload """
        try:
            nitori.unload_extension(f"plugins.{name}")
            nitori.load_extension(f"plugins.{name}")
        except Exception:
            await context.send(localization.get_string("pm_reload_error", context).format(plugin=name))
        await context.send(localization.get_string("pm_reload_success", context).format(plugin=name))

    @command(aliases=['exec', 'sh'])
    @permissions.system_administrators_only()
    async def shell(self, context, *, text: str):
        """ sys_help_shell """
        if not config.get("shell", False):
            return await context.send(localization.get_string("shell_disabled", context))
        message = await context.send(localization.get_string("shell_executing", context))
        async with context.channel.typing():
            result = await execute(text, pass_error=True)

        if not result:
            await message.delete()
            return await context.message.add_reaction('✅')

        if len(result) > 1500:
            data = BytesIO(result.encode('utf-8'))
            await message.delete()
            await context.send(content=localization.get_string("msg_length_limit", context),
                               file=File(data, filename=formatter.log_filename("output")))
        else:
            await message.edit(content=f"```fix\n{result}\n```")

    @command(aliases=['oauth', 'invite'])
    async def mkinvite(self, context):
        """ sys_help_mkinvite """
        await context.send(localization.get_string("invite_url", context).format(url=invite_url[0]))

    @command(aliases=['info'])
    async def about(self, context):
        """ sys_help_info """
        embed = Embed(color=Color.blue(),
                      title=localization.get_string("about_title", context).format(name=context.bot.user.name))
        embed.set_thumbnail(url=context.bot.user.avatar_url_as(format=None, static_format="png", size=256))
        embed.add_field(name=localization.get_string("homepage", context),
                        value=config.get("url", "http://localhost:7777/"),
                        inline=True)
        embed.add_field(
            name=localization.get_string("system_status", context),
            value=localization.get_string("status_operational", context) if redis_status() else
            localization.get_string("status_degraded", context),
            inline=True
        )
        embed.add_field(
            name=localization.get_string("software_license", context),
            value="GNU General Public License v3.0",
            inline=False
        )
        embed.add_field(
            name=localization.get_string("administrator_single", context) if len(system_administrators) == 1 else
            localization.get_string("administrator_multiple", context),
            value=', '.join([str(nitori.get_user(user)) for user in system_administrators]),
            inline=True)
        embed.add_field(
            name=localization.get_string("operator_single", context) if len(operators) == 1 else
            localization.get_string("operator_multiple", context),
            value=', '.join([str(nitori.get_user(user)) for user in operators]),
            inline=True)
        embed.set_footer(text="A Discord Utility by RandomChars",
                         icon_url="https://static.randomchars.net/img/RandomChars.png")
        await context.send(embed=embed)

    @command()
    @permissions.system_administrators_only()
    async def load(self, context):
        """ sys_help_load """
        embed = Embed(color=Color.blue(),
                      title="Load Information")
        embed.add_field(
            name="System Load",
            value=getloadavg(),
            inline=False
        )
        embed.add_field(
            name="CPU Usage",
            value=f"{cpu_percent()}%",
            inline=True
        )
        embed.add_field(
            name="Allocated Memory",
            value=f"{(Process(getpid()).memory_full_info().rss / 1024 ** 2):.2f} MiB",
            inline=True
        )
        embed.add_field(
            name="Total Memory",
            value=f"{(virtual_memory().total / 1024 / 1024 / 1024):.2f} GiB",
            inline=True
        )
        await context.send(embed=embed)

    @command()
    @guild_only()
    @permissions.operator_only()
    async def setexp(self, context, exp: int = None, user: Member = None):
        """ sys_help_setexp """
        if not exp:
            user = context.author
            experience.del_exp(user.id, context.guild.id)
            return await context.send(localization.get_string("exp_reset", context).format(name=user.display_name))
        if not user:
            user = context.author
            if exp == 0:
                experience.del_exp(user.id, context.guild.id)
                return await context.send(localization.get_string("exp_reset", context).format(name=user.display_name))
            experience.set_exp(user.id, context.guild.id, exp)
            return await context.send(
                localization.get_string("exp_set_e", context).format(name=user.display_name, exp=exp))
        if exp == 0:
            experience.del_exp(user.id, context.guild.id)
            return await context.send(localization.get_string("exp_reset", context).format(name=user.display_name))
        experience.set_exp(user.id, context.guild.id, exp)
        await context.send(localization.get_string("exp_set_e", context).format(name=user.display_name, exp=exp))

    @command()
    @guild_only()
    @permissions.operator_only()
    async def setlevel(self, context, level: int = None, user: Member = None):
        """ sys_help_setlevel """
        if not level:
            user = context.author
            experience.del_exp(user.id, context.guild.id)
            return await context.send(localization.get_string("exp_reset", context).format(name=user.display_name))
        if not user:
            user = context.author
            if level == 0:
                experience.del_exp(user.id, context.guild.id)
                return await context.send(localization.get_string("exp_reset", context).format(name=user.display_name))
            exp = experience.calculate_exp(level)
            experience.set_exp(user.id, context.guild.id, exp)
            return await context.send(
                localization.get_string("exp_set_l", context).format(name=user.display_name, level=level))
        if level == 0:
            experience.del_exp(user.id, context.guild.id)
            return await context.send(localization.get_string("exp_reset", context).format(name=user.display_name))
        exp = experience.calculate_exp(level)
        experience.set_exp(user.id, context.guild.id, exp)
        await context.send(localization.get_string("exp_set_l", context).format(name=user.display_name, level=level))

    @command()
    @permissions.operator_only()
    async def echo(self, context, *, message):
        """ sys_help_echo """
        if permissions.manage(context):
            await context.message.delete()
        async with context.channel.typing():
            await sleep(delay=len(str(message)) * 0.1)
            await context.send(message)


def setup(client):
    client.add_cog(System())
