""" Get and set customizations. """

from base64 import b64encode, b64decode
from discord import TextChannel, Role
from discord.utils import get
from aya import nitori, redis, redis_status, config
from aya.utils import localization


def get_entry(name, key):
    if redis_status():
        result = redis.hget(name, key)
        if result:
            return result.decode("utf-8")
    return None


def get_all_keys(key):
    origin = redis.hgetall(key)
    result = {}
    for key_encoded in origin:
        try:
            key = int(key_encoded.decode("utf-8"))
            result.update({key: int(origin[key_encoded].decode("utf-8"))})
        except ValueError:
            key = key_encoded.decode("utf-8")
            result.update({key: origin[key_encoded].decode("utf-8")})
    if origin is None:
        return None
    return result


def reset_guild(gid):
    redis.delete(f"settings.{gid}")
    redis.delete(f"exp.{gid}")
    redis.delete(f"rank.{gid}")
    redis.delete(f"exp_bl.{gid}")
    redis.delete(f"ra_metadata.{gid}")
    redis.delete(f"ra_table_0.{gid}")
    redis.delete(f"ra_table_1.{gid}")
    redis.delete(f"ra_table_2.{gid}")
    redis.delete(f"ra_table_3.{gid}")
    redis.delete(f"ra_table_4.{gid}")
    redis.delete(f"ra_table_5.{gid}")
    redis.delete(f"ra_table_6.{gid}")
    redis.delete(f"ra_table_7.{gid}")


def get_message(gid, message):
    binary_message = get_entry(f"settings.{gid}", f"message.{message}")
    return b64decode(binary_message.encode("utf-8")).decode("utf-8") \
        if binary_message else None


def set_message(gid, message, content):
    if message == "" or message is None:
        reset_message(gid, message)
        return
    redis.hset(f"settings.{gid}", f"message.{message}", b64encode(content.encode("utf-8")))


def reset_message(gid, message):
    redis.hdel(f"settings.{gid}", f"message.{message}")


def get_prefix(gid):
    prefix_binary = get_entry(f"settings.{gid}", "prefix")
    return b64decode(prefix_binary.encode("utf-8")).decode("utf-8") \
        if prefix_binary else False


def set_prefix(gid, content):
    if content == "" or content is None:
        reset_prefix(gid)
        return
    redis.hset(f"settings.{gid}", "prefix", b64encode(content.encode("utf-8")))


def reset_prefix(gid):
    redis.hdel(f"settings.{gid}", "prefix")


def get_language(gid):
    locale = get_entry(f"settings.{gid}", "language")
    return locale if locale else config["default_language"]


def set_language(gid, lang):
    if lang not in localization.locales:
        raise RuntimeWarning(f"{lang} was not implemented, an error occurred in the configurator?")
    redis.hset(f"settings.{gid}", "language", lang)


def get_ranks(gid):
    return get_all_keys(f"rank.{gid}")


def get_rank(gid, level):
    return get_entry(f"rank.{gid}", f"{level}")


def set_rank(gid, level, rid):
    redis.hset(f"rank.{gid}", f"{level}", int(rid))


def remove_rank(gid, level):
    redis.hdel(f"rank.{gid}", f"{level}")


def get_exp_bls(gid):
    return get_all_keys(f"exp_bl.{gid}")


def get_exp_bl(gid, channel):
    return get_entry(f"exp_bl.{gid}", channel) == "1"


def set_exp_bl(gid, channel):
    redis.hset(f"exp_bl.{gid}", channel, 1)


def remove_exp_bl(gid, channel):
    redis.hdel(f"exp_bl.{gid}", channel)


def get_assignable_role_pages(gid):
    pages = get_entry(f"ra_metadata.{gid}", "total_pages")
    if not pages:
        return 0
    return int(pages)


def create_assignable_role_page(gid, name):
    page = get_assignable_role_pages(gid)
    if page > 7:
        raise ValueError("Going above 8 tables is not allowed.")
    redis.hset(f"ra_metadata.{gid}", "total_pages", page + 1)
    if page == 0 and redis.exists(f"assignable_role.{gid}"):
        redis.rename(f"assignable_role.{gid}", f"ra_table_0.{gid}")
        database_record = get_message(gid, "role_assign_message")
        if database_record:
            message_id, channel_id = database_record.split(' ', 1)
            channel_id = int(channel_id)
            message_id = int(message_id)
            set_assignable_role_message_info(gid, page, channel_id, message_id)
            reset_message(gid, "role_assign_message")
    set_assignable_role_page_title(gid, page, name)
    return page


def drop_assignable_role_page(gid, page):
    pages = get_assignable_role_pages(gid)
    if pages == 0:
        raise ValueError("No pages left.")
    if page > pages - 1:
        raise ValueError("Out of bounds page.")
    redis.hset(f"ra_metadata.{gid}", "total_pages", pages - 1)
    if page == pages - 1:
        redis.delete(f"ra_table_{page}.{gid}")
    else:
        redis.delete(f"ra_table_{page}.{gid}")
        for move_page in range(page + 1, pages, 1):
            redis.renamenx(f"ra_table_{move_page}.{gid}", f"ra_table_{move_page - 1}.{gid}")


def set_assignable_role_page_title(gid, page, content):
    if page > 7:
        raise ValueError("Going above 8 tables is not allowed.")
    redis.hset(f"ra_table_{page}.{gid}", "m_title", b64encode(content.encode("utf-8")))


def get_assignable_role_page_title(gid, page):
    if page > 7:
        raise ValueError("Going above 8 tables is not allowed.")
    return b64decode(get_entry(f"ra_table_{page}.{gid}", "m_title")).decode("utf-8")


def get_assignable_role_message_info(gid, page):
    if page > 7:
        raise ValueError("Going above 8 tables is not allowed.")
    channel_id = get_entry(f"ra_table_{page}.{gid}", "m_channel")
    message_id = get_entry(f"ra_table_{page}.{gid}", "m_message")
    if channel_id is None or message_id is None:
        return None, None
    return int(channel_id), int(message_id)


def get_assignable_role_messages(gid):
    pages = get_assignable_role_pages(gid)
    if pages == 0:
        return None, None
    channel_ids = {}
    message_ids = {}
    for page in range(0, pages, 1):
        channel_id, message_id = get_assignable_role_message_info(gid, page)
        channel_ids.update({channel_id: page})
        message_ids.update({message_id: page})
    return channel_ids, message_ids


def set_assignable_role_message_info(gid, page, channel, message):
    if page > 7:
        raise ValueError("Going above 8 tables is not allowed.")
    redis.hset(f"ra_table_{page}.{gid}", "m_channel", channel)
    redis.hset(f"ra_table_{page}.{gid}", "m_message", message)


def get_assignable_roles(gid, page):
    page_data = get_all_keys(f"ra_table_{page}.{gid}")
    result = {}
    for key in page_data:
        if not str(key).startswith("m_"):
            result.update({key: page_data[key]})
    return result


def get_assignable_role(gid, page, emote):
    return get_entry(f"ra_table_{page}.{gid}", emote)


def set_assignable_role(gid, page, role, emote):
    redis.hset(f"ra_table_{page}.{gid}", emote, role)


def remove_assignable_role(gid, page, emote):
    redis.hdel(f"ra_table_{page}.{gid}", emote)


def set_log_channel(gid, channel: TextChannel):
    channel_id = channel.id
    redis.hset(f"settings.{gid}", "logging", channel_id)


def get_log_channel(gid):
    channel_id = get_entry(f"settings.{gid}", "logging")
    if channel_id is None:
        return None
    return nitori.get_channel(int(channel_id))


def remove_log_channel(gid):
    redis.hdel(f"settings.{gid}", "logging")


def set_gulag_channel(gid, channel: TextChannel):
    channel_id = channel.id
    redis.hset(f"settings.{gid}", "gulag", channel_id)


def get_gulag_channel(gid):
    channel_id = get_entry(f"settings.{gid}", "gulag")
    if channel_id is None:
        return None
    return nitori.get_channel(int(channel_id))


def remove_gulag_channel(gid):
    redis.hdel(f"settings.{gid}", "gulag")


def set_mute_role(gid, role: Role):
    role_id = role.id
    redis.hset(f"settings.{gid}", "mute_role", role_id)


def get_mute_role(gid):
    role_id = get_entry(f"settings.{gid}", "mute_role")
    if role_id is None:
        return None
    return get(nitori.get_guild(gid).roles, id=int(role_id))


def remove_mute_role(gid):
    redis.hdel(f"settings.{gid}", "mute_role")


def get_slur_filter_status(gid):
    thing = get_entry(f"settings.{gid}", "slur_filter")
    if thing is None:
        thing = "0"
    return bool(int(thing))


def toggle_slur_filter_status(gid):
    thing = get_slur_filter_status(gid)
    if thing:
        redis.hset(f"settings.{gid}", "slur_filter", "0")
        return False
    else:
        redis.hset(f"settings.{gid}", "slur_filter", "1")
        return True
